@@ -5,7 +5,7 @@ import time
 from dataclasses import dataclass
 from datetime import datetime
 from pathlib import Path
-from typing import Any, Iterable, Mapping, MutableMapping, Sequence
+from typing import Any, Iterable, Mapping, MutableMapping, Sequence, cast
 from urllib.parse import parse_qsl, urlencode, urlsplit
 
 import duckdb
@@ -55,6 +55,7 @@ class RouteExecutionResult:
     total_rows: int
     offset: int
     limit: int | None
+    meta: Mapping[str, object] | None
 
 
 @dataclass(slots=True, frozen=True)
@@ -122,7 +123,7 @@ def create_app(routes: Sequence[RouteDefinition], config: Config) -> FastAPI:
     if not routes:
         raise ValueError("At least one route must be provided to create the application")
 
-    app = FastAPI(title="webbed_duck", version="0.4.3")
+    app = FastAPI(title="webbed_duck", version="0.4.6")
     app.state.config = config
     app.state.analytics = AnalyticsStore(
         weight=config.analytics.weight_interactions,
@@ -167,9 +168,9 @@ def create_app(routes: Sequence[RouteDefinition], config: Config) -> FastAPI:
     if config.auth.mode == "pseudo":
         @app.post("/auth/pseudo/session")
         async def create_pseudo_session(request: Request) -> Mapping[str, object]:
-            payload = await request.json()
-            if not isinstance(payload, Mapping):
-                raise _http_error("invalid_parameter", "Session payload must be an object")
+            payload = _require_json_mapping(
+                await request.json(), message="Session payload must be an object"
+            )
             email_raw = payload.get("email")
             remember = bool(payload.get("remember_me", False))
             try:
@@ -275,9 +276,9 @@ def create_app(routes: Sequence[RouteDefinition], config: Config) -> FastAPI:
         override_meta = metadata.get("overrides", {}) if isinstance(metadata, Mapping) else {}
         allowed = set(_coerce_sequence(override_meta.get("allowed")))
         key_columns = _coerce_sequence(override_meta.get("key_columns"))
-        payload = await request.json()
-        if not isinstance(payload, Mapping):
-            raise _http_error("invalid_parameter", "Override payload must be an object")
+        payload = _require_json_mapping(
+            await request.json(), message="Override payload must be an object"
+        )
         column = str(payload.get("column", "")).strip()
         if not column:
             raise _http_error("invalid_parameter", "Override column is required")
@@ -318,9 +319,9 @@ def create_app(routes: Sequence[RouteDefinition], config: Config) -> FastAPI:
         if not columns:
             raise HTTPException(status_code=500, detail={"code": "append_misconfigured", "message": "Append metadata must declare columns"})
         destination = str(append_meta.get("destination") or f"{route.id}.csv")
-        payload = await request.json()
-        if not isinstance(payload, Mapping):
-            raise _http_error("invalid_parameter", "Append payload must be an object")
+        payload = _require_json_mapping(
+            await request.json(), message="Append payload must be an object"
+        )
         record = {column: payload.get(column) for column in columns}
         try:
             path = append_record(
@@ -342,9 +343,9 @@ def create_app(routes: Sequence[RouteDefinition], config: Config) -> FastAPI:
         user = await app.state.auth_adapter.authenticate(request)
         if not user:
             raise HTTPException(status_code=401, detail={"code": "not_authenticated", "message": "Login required to create shares"})
-        payload = await request.json()
-        if not isinstance(payload, Mapping):
-            raise _http_error("invalid_parameter", "Share payload must be an object")
+        payload = _require_json_mapping(
+            await request.json(), message="Share payload must be an object"
+        )
         default_fmt = route.default_format or "html_t"
         fmt_raw = payload.get("format")
         fmt = fmt_raw.lower() if isinstance(fmt_raw, str) else default_fmt.lower()
@@ -621,6 +622,7 @@ def _render_route_response(
             total_rows=result.total_rows,
             offset=result.offset,
             limit=result.limit,
+            meta=result.meta,
         )
 
     metadata = route.metadata if isinstance(route.metadata, Mapping) else {}
@@ -745,6 +747,7 @@ def _execute_route(
         total_rows=cache_result.total_rows,
         offset=cache_result.applied_offset,
         limit=applied_limit,
+        meta=cache_result.meta,
     )
 
 
@@ -795,6 +798,7 @@ def _format_response(
             format_hint=fmt,
             pagination=pagination_values,
             rpc_payload=rpc_payload,
+            cache_meta=result.meta,
         )
         response = HTMLResponse(html)
         _attach_rpc_headers(response, result)
@@ -819,6 +823,7 @@ def _format_response(
             format_hint=fmt,
             pagination=pagination_values,
             rpc_payload=rpc_payload,
+            cache_meta=result.meta,
         )
         response = HTMLResponse(html)
         _attach_rpc_headers(response, result)
@@ -1108,6 +1113,7 @@ def _build_share_artifacts(
             charts_meta,
             postprocess=post_opts,
             watermark=watermark_text,
+            cache_meta=result.meta,
         )
 
     attachments: list[tuple[str, bytes]] = []
@@ -1130,6 +1136,7 @@ def _build_share_artifacts(
                     charts_meta,
                     postprocess=post_opts,
                     watermark=watermark_text,
+                    cache_meta=result.meta,
                 )
             attachments.append((f"{route.id}.html", html_body.encode("utf-8")))
 
@@ -1202,9 +1209,10 @@ def _render_share_email(
 
 def _collect_params(route: RouteDefinition, request: Request) -> Mapping[str, object]:
     values: dict[str, object] = {}
+    query_params = request.query_params
     for spec in route.params:
-        raw_value = request.query_params.get(spec.name)
-        if raw_value is None:
+        raw_values = query_params.getlist(spec.name)
+        if not raw_values:
             if spec.default is not None:
                 values[spec.name] = spec.default
             elif spec.required:
@@ -1212,10 +1220,21 @@ def _collect_params(route: RouteDefinition, request: Request) -> Mapping[str, ob
             else:
                 values[spec.name] = None
             continue
+        if len(raw_values) == 1:
+            raw_value = raw_values[0]
+            try:
+                values[spec.name] = spec.convert(raw_value)
+            except ValueError as exc:
+                raise _http_error("invalid_parameter", str(exc)) from exc
+            continue
+
+        converted_values: list[object] = []
         try:
-            values[spec.name] = spec.convert(raw_value)
+            for raw_value in raw_values:
+                converted_values.append(spec.convert(raw_value))
         except ValueError as exc:
             raise _http_error("invalid_parameter", str(exc)) from exc
+        values[spec.name] = converted_values
     return values
 
 
@@ -1430,6 +1449,12 @@ def _get_route(routes: Sequence[RouteDefinition], route_id: str) -> RouteDefinit
     raise HTTPException(status_code=404, detail={"code": "not_found", "message": f"Route '{route_id}' not found"})
 
 
+def _require_json_mapping(payload: object, *, message: str) -> Mapping[str, object]:
+    if not isinstance(payload, Mapping):
+        raise _http_error("invalid_parameter", message)
+    return cast(Mapping[str, object], payload)
+
+
 def _http_error(code: str, message: str) -> HTTPException:
     entry = _ERROR_TAXONOMY.get(code, {})
     detail = {"code": code, "message": message}
